%{
/* This file is part of mailfromd.
   Copyright (C) 2005, 2006, 2007 Sergey Poznyakoff

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
   MA 02110-1301 USA */

#define MF_SOURCE_NAME  MF_SOURCE_LEX
	
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif
#include <sys/types.h>	
#include <sys/stat.h>
#define obstack_chunk_alloc malloc
#define obstack_chunk_free free
#include <obstack.h>
#include "mailfromd.h"
#include "gram.h"
#include "prog.h"
	
struct input_file_ident {
	ino_t i_node;
	dev_t device;
};
 
static struct locus locus; /* Current input location */
static struct input_file_ident source_id; /* Identification of the input
					     file */
 
static struct obstack string_stk; /* Obstack for constructing string values */
static char *multiline_delimiter; /* End of here-document delimiter */ 
static size_t multiline_delimiter_len; /* Length of multiline_delimiter_len */
static int multiline_unescape;         /* Unescape here-document contents */
static int (*char_to_strip)(char);     /* Strip matching characters of each
					  here-document line */

static int
is_tab(char c)
{
	return c == '\t';
}
 
static int
is_space(char c)
{
	return c == '\t' || c == ' ';
}

#define line_begin string_begin
#define line_add string_add
#define line_add_char string_add_char
static void line_finish(void);
static void string(const char *str, size_t len);
static int isemptystr(char *text);
static int builtin_macro(const char *s);

/* External interface to locus */ 
const struct locus *
get_locus()
{
	return &locus;
}

/* Auxiliary function for returning keyword tokens */ 
static int
keyword(int kw)
{
	yylval.locus = locus;
	return kw;
}


/* Input stack support */

#define xinput() (yyin ? getc(yyin) : EOF)

#undef YY_INPUT
#define YY_INPUT(buf,result,max_size)  do {                             \
        int i;                                                          \
        for (i = 0; i < max_size; i++) {                                \
                int ch = xinput();                                      \
                if (ch == EOF)                                          \
                        break;                                          \
                buf[i] = ch;                                            \
        }                                                               \
        result = i;                                                     \
} while (0)

#define LEX_BUFFER_STATE YY_BUFFER_STATE

#define SET_BUFFER_STATE(s) do {                                        \
        (s) = YY_CURRENT_BUFFER;                                        \
        yy_switch_to_buffer(yy_create_buffer(yyin, YY_BUF_SIZE));       \
} while (0)

#define RESTORE_BUFFER_STATE(s) do {                                    \
        yy_delete_buffer(YY_CURRENT_BUFFER);                            \
        yy_switch_to_buffer(s);                                         \
} while (0)

struct buffer_ctx {
	struct buffer_ctx *prev;
	struct locus locus;
	struct input_file_ident id;
	FILE *yyin;
	LEX_BUFFER_STATE state;
};
	
static struct buffer_ctx *context_stack;

#define STAT_ID_EQ(st,id) ((id).i_node == (st).st_ino       \
			   && (id).device == (st).st_dev)

struct buffer_ctx *
ctx_lookup(struct stat *st)
{
	struct buffer_ctx *ctx;

	for (ctx = context_stack; ctx; ctx = ctx->prev)
		if (STAT_ID_EQ(*st, ctx->id))
			break;
	return ctx;
}

static mu_list_t include_path;
static mu_list_t std_include_path;

struct file_data {
	const char *name;
	size_t namelen;
	char *buf;
	size_t buflen;
	int found;
};

static int
find_include_file(void *item, void *data)
{
	char *dir = item;
	struct file_data *dptr = data;
	size_t size = strlen(dir) + 1 + dptr->namelen + 1;
	if (size > dptr->buflen) {
		dptr->buflen = size;
		dptr->buf = xrealloc(dptr->buf, dptr->buflen);
	}
	strcpy(dptr->buf, dir);
	strcat(dptr->buf, "/");
	strcat(dptr->buf, dptr->name);
	return dptr->found = access(dptr->buf, F_OK) == 0;
}

void
lex_setup()
{
	mu_list_create(&include_path);
	mu_list_create(&std_include_path);
	mu_list_append(std_include_path, DEFAULT_VERSION_INCLUDE_DIR);
	mu_list_append(std_include_path, DEFAULT_INCLUDE_DIR);
	mu_list_append(std_include_path, "/usr/share/mailfromd/include");
	mu_list_append(std_include_path, "/usr/local/share/mailfromd/include");
}

static mu_list_t /* of struct input_file_ident */ incl_sources;

static int
input_file_ident_cmp(const void *item, const void *data)
{
	const struct input_file_ident *id = item;
	const struct stat *st = data;
	return !STAT_ID_EQ(*st, *id);
}

static void
input_file_ident_free(void *item)
{
	free(item);
}

static int
source_lookup(struct stat *st)
{
	int rc;
	if (!incl_sources) {
		mu_list_create(&incl_sources);
		mu_list_set_comparator(incl_sources, input_file_ident_cmp);
		mu_list_set_destroy_item(incl_sources, input_file_ident_free);
	}
	rc = mu_list_locate (incl_sources, st, NULL);
	if (rc == MU_ERR_NOENT) {
		struct input_file_ident *id = xmalloc(sizeof *id);
		id->i_node = st->st_ino;
		id->device = st->st_dev;
		mu_list_append(incl_sources, id);
	}
	return rc == 0;
}

int
push_source(const char *name, int once)
{
	FILE *fp;
	struct buffer_ctx *ctx;
	struct stat st;
	struct literal *lit;
	
	if (stat(name, &st)) {
		parse_error(_("Cannot stat `%s': %s"), name, mu_strerror(errno));
		return 1;
	}

	if (locus.file && STAT_ID_EQ(st, source_id)) {
		parse_error(_("Recursive inclusion"));
		return 1;
	}

	if ((ctx = ctx_lookup(&st))) {
		parse_error(_("Recursive inclusion"));
		if (ctx->prev)
			parse_error_locus(&ctx->prev->locus,
					  _("`%s' already included here"),
					  name);
		else
			parse_error(_("`%s' already included at top level"),
				    name);
		return 1;
	}

	if (once && source_lookup(&st))
		return 0;
	
	fp = fopen(name, "r");
	if (!fp) {
		parse_error(_("Cannot open `%s': %s"), name, mu_strerror(errno));
		return 1;
	}

	/* Push current context */
	if (locus.file) {
		ctx = xmalloc (sizeof (*ctx));
		ctx->locus = locus;
		ctx->id = source_id;
		ctx->yyin = yyin;
		ctx->prev = context_stack;
		context_stack = ctx;
      
		/* Switch to the new context */
		yyin = fp;
		SET_BUFFER_STATE (ctx->state);
	} else {
		yyrestart (fp);
	}
	lit = string_alloc(name, strlen(name));
	lit->flags |= VAR_REFERENCED;
	locus.file = lit->text;
	locus.line = 1;
	source_id.i_node = st.st_ino;
	source_id.device = st.st_dev;

	if (yy_flex_debug)
		fprintf(stderr, "Processing file `%s'\n", name);

	return 0;
}

int
pop_source()
{
	struct buffer_ctx *ctx;

	if (yyin)
		fclose(yyin);
	if (!context_stack) {
		yyin = NULL;
		locus.file = NULL;
		return 1;
	}
	/* Restore previous context */
	locus = context_stack->locus;
	if (yy_flex_debug) 
		fprintf(stderr, "Resuming file `%s' at line %lu\n",
			locus.file, (unsigned long) locus.line);
	source_id = context_stack->id;
	RESTORE_BUFFER_STATE(context_stack->state);
	ctx = context_stack->prev;
	free(context_stack);
	context_stack = ctx;

	return 0;
}

int
try_file(const char *name, int allow_cwd, char **newp)
{
	static char *cwd = ".";
	struct file_data fd;
			
	fd.name = name;
	fd.namelen = strlen(name);
	fd.buf = NULL;
	fd.buflen = 0;
	fd.found = 0;
	
	if (allow_cwd) {
		mu_list_prepend(include_path, cwd);
		mu_list_do(include_path, find_include_file, &fd);
		mu_list_remove(std_include_path, cwd);
	} else
		mu_list_do(include_path, find_include_file, &fd);
			
	if (!fd.found) {
		mu_list_do(std_include_path, find_include_file, &fd);
					
		if (!fd.found) {
			parse_error(_("%s: No such file or directory"), name);
			*newp = NULL;
		} 
	}
	if (fd.found)
		*newp = fd.buf;
	return fd.found;
}

void
parse_include(int once)
{
	int argc;
	char **argv;
	char *tmp = NULL;
	char *p = NULL;
	
	if (mu_argcv_get(yytext, "", NULL, &argc, &argv)) 
		parse_error(_("Cannot parse include line"));
	else if (argc != 2) 
		parse_error(_("Invalid include statement"));
	else {
		size_t len;
		int allow_cwd;

		p = argv[1];
		len = strlen(p);
		
		if (p[0] == '<' && p[len - 1] == '>') {
			allow_cwd = 0;
			p[len - 1] = 0;
			p++;
		} else 
			allow_cwd = 1;

		if (p[0] != '/' && try_file(p, allow_cwd, &tmp))
			p = tmp;
	}

	locus.line++;
	if (p)
		push_source(p, once);
	free(tmp);
	mu_argcv_free(argc, argv);
}

#define DEFAULT_SUFFIX ".mf"

void
parse_require()
{
	int argc;
	char **argv;
	char *tmp = NULL;
	char *p = NULL;
	
	if (mu_argcv_get(yytext, "", NULL, &argc, &argv)) 
		parse_error(_("Cannot parse require line"));
	else if (argc != 2) 
		parse_error(_("Invalid require statement"));
	else {
		size_t len;
		char *fname;
		
		p = argv[1];
		len = strlen(p);

		/* For compatibility with #include */
		if (p[0] == '<' && p[len - 1] == '>') {
			p[--len] = 0;
			p++;
		} 

		if (len < sizeof DEFAULT_SUFFIX
		    || memcmp(p + len - sizeof DEFAULT_SUFFIX - 1,
			      DEFAULT_SUFFIX,
			      sizeof DEFAULT_SUFFIX - 1)) {
			fname = xmalloc(len + sizeof DEFAULT_SUFFIX);
			strcpy(fname, p);
			strcat(fname, DEFAULT_SUFFIX);
			p = fname;
		} else
			fname = NULL;
			
		if (p[0] != '/') {
			if (try_file(p, 0, &tmp))
				p = tmp;
			else
				p = NULL;
			free(fname);
		} else if (fname) {
			tmp = fname;
			fname = NULL;
		}
	}
	locus.line++;
	if (p)
		push_source(p, 1);
	free(tmp);
	mu_argcv_free(argc, argv);
}

static int
get_const(const char *name)
{
	struct value *value_ptr;
	if (value_ptr = constant_lookup(name)) {
		switch (value_ptr->type) {
		case dtype_number:
			yylval.number = value_ptr->v.number;
			return NUMBER;

		case dtype_string:
			yylval.literal = value_ptr->v.literal;
			return STRING;
			
		default:
			abort();
		}
	}
	return 0;
}
		
%}

/* Exclusive states:

   COMMENT        Within a C-style comment;
   STR            Processing a complex string;
   ML             Within a multi-line aggregator.  The line being built
                  requires stripping leading whitespace (if requested).
   CML            Continuation within a multi-line aggregator.  The line
                  being built does not require stripping leading whitespace.
   QML            Quoted multi-line aggregator.  No variable substitution and
                  unquoting is needed.

   Inclusive states:
   
   ONBLOCK	  Lexical tie-in after an `on' keyword.  In ONBLOCK state
                  the strigns `as', `host', `for', `from', and `poll' are
		  recognized as keywords.    	   
*/

%x COMMENT STR ML CML QML
%s ONBLOCK

N [0-9][0-9]*
P [1-9][0-9]*
X [0-9a-fA-F]
O [0-7]
WS [ \t][ \t]*
IDENT [a-zA-Z_][a-zA-Z_0-9]*
LOCUS __file__|__line__|__function__
VMACRO __package__|__version__|__major__|__minor__|__patch__
MACRO {LOCUS}|{VMACRO}|__statedir__
%%
         /* C-style comments */
"/*"         BEGIN(COMMENT);
<COMMENT>[^*\n]*        /* eat anything that's not a '*' */
<COMMENT>"*"+[^*/\n]*   /* eat up '*'s not followed by '/'s */
<COMMENT>\n             ++locus.line;
<COMMENT>"*"+"/"        BEGIN(INITIAL);
         /* Configuration directives */
^[ \t]*#[ \t]*pragma[ \t].*/\n parse_pragma(yytext);
^[ \t]*#[ \t]*include_once[ \t].*\n parse_include(1);
^[ \t]*#[ \t]*include[ \t].*\n parse_include(0);
^[ \t]*#[ \t]*require[ \t].*\n parse_require(yytext);
^[ \t]*#[ \t]*error[ \t].*\n  { parse_error("%s", yytext); locus.line++; }
         /* End-of-line comments */
#.*\n   { locus.line++; }
#.*     /* end-of-file comment */;
         /* Reserved words */
accept    return keyword(ACT_ACCEPT);
reject    return keyword(ACT_REJECT);
tempfail  return keyword(ACT_TEMPFAIL);
continue  return keyword(ACT_CONTINUE);
discard   return keyword(ACT_DISCARD);
add       return keyword(ADD);
replace   return keyword(REPLACE);
delete    return keyword(DELETE);
if        return keyword(IF);
fi        return keyword(FI);
else      return keyword(ELSE);
elif      return keyword(ELIF);
on        return keyword(ON);
do        return keyword(DO);
done      return keyword(DONE);
matches   return keyword(MATCHES);
fnmatches return keyword(FNMATCHES);
mx{WS}matches return keyword(MXMATCHES);
mx{WS}fnmatches return keyword(MXFNMATCHES);
when      return keyword(WHEN);
or        return keyword(OR);
and       return keyword(AND);
not       return keyword(NOT);
next      return keyword(NEXT);
prog      return keyword(PROG);
set       return keyword(SET);
catch     return keyword(CATCH);
echo      return keyword(KW_ECHO);
return    return keyword(RETURN);
returns   return keyword(RETURNS);
func      return keyword(FUNC);
switch    return keyword(SWITCH);
case      return keyword(CASE);
default   return keyword(DEFAULT);
string    { yylval.type = dtype_string; return TYPE; }
number    { yylval.type = dtype_number; return TYPE; }
const     return keyword(CONST);
throw     return keyword(THROW);
loop      return keyword(LOOP);
while     return keyword(WHILE);
for       return keyword(FOR);
break     return keyword(BREAK);
pass      return keyword(PASS);
begin     return keyword(KW_BEGIN);
end       return keyword(KW_END);

{MACRO}   { return builtin_macro(yytext); } 
<ONBLOCK>poll      return keyword(POLL);
<ONBLOCK>host      return keyword(HOST);
<ONBLOCK>as        return keyword(AS);
<ONBLOCK>from      return keyword(FROM);
         /* Variables */
<INITIAL,ONBLOCK,ML,CML,STR>\%({MACRO}) { return builtin_macro(yytext+1); }
<INITIAL,ONBLOCK,ML,CML,STR>\%{IDENT}     {
	int rc;
	string(yytext + 1, yyleng - 1);
	if ((rc = get_const(yylval.literal->text)) != 0)
		return rc;
	else
		return VARIABLE; }
<INITIAL,ONBLOCK,ML,CML,STR>\%\{{IDENT}\} {
	int rc;
	string(yytext + 2, yyleng - 3);
	if ((rc = get_const(yylval.literal->text)) != 0)
		return rc;	
	else
		return VARIABLE; }
         /* Positional arguments */
<INITIAL,ONBLOCK,ML,CML,STR>\${P}         {
	yylval.number = strtol(yytext + 1, NULL, 0); return ARG; }
         /* Sendmail variables */
<INITIAL,ONBLOCK,ML,CML,STR>\${IDENT} {
	        if (yyleng == 2)
	           string(yytext + 1, 1);
                else {
                   line_begin();
                   line_add("{", 1);
                   line_add(yytext + 1, yyleng - 1);
                   line_add("}", 1);
                   line_finish();
                }
                return SYMBOL;
              }
<INITIAL,ONBLOCK,ML,CML,STR>\$\{{IDENT}\} {
	string(yytext+1, yyleng - 1); return SYMBOL; }
         /* Exception conditions */
&{IDENT}      { int rc;
                mf_status status;
		static int warned_once;
		
                /*DEPRECATION*/ 
                if ((rc = get_const(yytext + 1)) != 0) {
			if (!warned_once) {
				parse_warning("Obsolete constant "
					      "form used: %s", yytext);
				parse_warning("Remove leading '&'");
				warned_once = 1;
			}
			return rc;
		} else if (string_to_stat(yytext + 1, &status)) {
			parse_warning(_("Unknown exception code: %s"),
			              yytext + 1);
			string(yytext + 1, yyleng - 1);
			return IDENTIFIER;
		} else {
			if (!warned_once) {
				parse_warning("Obsolete constant "
					      "form used: %s", yytext);
				parse_warning("Remove leading '&' "
					      "and include <status.mfh>");
				yylval.number = status;
				warned_once = 1;
			}
			return NUMBER;
		}
         }
         /* Back-references */
<INITIAL,ONBLOCK,ML,CML,STR>\\{P} {
	 yylval.number = strtoul(yytext+1, NULL, 0);
	 return BACKREF; }
         /* Numeric strings */
{N}\.{N}\.{N} {	string(yytext, yyleng); return XCODE; }
[0-9]{3}      { string(yytext, yyleng); return CODE; }
0[xX]{X}{X}*  { yylval.number = strtoul(yytext, NULL, 16); return NUMBER; };
0{O}{O}*      { yylval.number = strtoul(yytext, NULL, 8); return NUMBER; };
0|{P}         { yylval.number = strtoul(yytext, NULL, 10); return NUMBER; };
         /* Strings */
{IDENT}  {
	int paren_follows = 0;
	enum { is_ident, is_builtin, is_func } state;

	if (yylval.builtin = builtin_lookup(yytext))
		state = is_builtin;
	else if (yylval.function = function_lookup(yytext))
		state = is_func;
	else
		state = is_ident;

	if (state != is_ident) {
		int c;
		
		while ((c = input()) && isascii(c) && isspace(c))
			if (c == '\n')
				++locus.line;
	
		paren_follows = (c == '(');
		unput(c);
	}
	
	if (state == is_builtin) 
		return yylval.builtin->rettype == dtype_unspecified ?
			BUILTIN_PROC :
			(paren_follows ? BUILTIN : BUILTIN_P);
	else if (state == is_func)
		return yylval.function->rettype == dtype_unspecified ?
			FUNCTION_PROC :
			(paren_follows ? FUNCTION : FUNCTION_P);
	else {
		string(yytext, yyleng);
		return IDENTIFIER;
	}
  }
'[^\n']*'             { string(yytext+1, yyleng-2); return STRING; }
\"[^\\\"$%\n]*\"      { string(yytext+1, yyleng-2); return STRING; }
\"[^\\\"$%\n]*\\\n    { ++locus.line;
                        BEGIN(STR);
                        line_begin();
                        line_add(yytext + 1, yyleng - 3); }     
\"[^\\\"$%\n]*/[\\$%] { BEGIN(STR);
                        string(yytext+1, yyleng-1);
                        line_begin();
                        return STRING; }
\"\\x{X}{X}/[\\$%] {
         BEGIN(STR);
         line_add_char(strtoul(yytext + 3, NULL, 16));
         line_finish();
         return STRING;
}
\"\\x{X}{X} {
         BEGIN(STR);
         line_add_char(strtoul(yytext + 3, NULL, 16));
}

\"\\0{O}{O}{O}/[\\$%] {
         BEGIN(STR);  
         line_add_char(strtoul(yytext + 3, NULL, 8));
         line_finish();
         return STRING;
}
\"\\0{O}{O}{O} {
         BEGIN(STR);  
         line_add_char(strtoul(yytext + 3, NULL, 8));
}

\"\\[^1-9]/[\\$%]  {
         BEGIN(STR);
         line_add_char(mu_argcv_unquote_char(yytext[2]));
         line_finish();
         return STRING;
}
\"\\[^1-9]  {
         BEGIN(STR);
         line_add_char(mu_argcv_unquote_char(yytext[2]));
}
<STR>[^\\\"$%\n]*\\\n { ++locus.line; line_add(yytext, yyleng - 2); }
<STR>[^\\\"$%\n]*\"   { BEGIN(INITIAL);
                        if (yyleng > 1) 
                          line_add(yytext, yyleng - 1); 
                        line_finish();
		        return STRING; }
<STR>[^\\\"$%\n]+/[\\$%] {
                       line_add(yytext, yyleng);
                       line_finish();
                       line_begin();
                       return STRING;
                 }
<STR,ML,CML>\\x{X}{X}/[\\$%] {
         line_add_char(strtoul(yytext + 2, NULL, 16));
         line_finish();
         line_begin();
         return STRING;
}
<STR,ML,CML>\\x{X}{X} {
         line_add_char(strtoul(yytext + 2, NULL, 16));
}
<STR,ML,CML>\\0{O}{O}{O}/[\\$%] {
         line_add_char(strtoul(yytext + 2, NULL, 8));
         line_finish();
         line_begin();
         return STRING;
}
<STR,ML,CML>\\0{O}{O}{O} {
         line_add_char(strtoul(yytext + 2, NULL, 8));
}
<STR,ML,CML>\\[^1-9]/[\\$%] {
         line_add_char(mu_argcv_unquote_char(yytext[1]));
         line_finish();
         line_begin();
         return STRING;
}
<STR,ML,CML>\\[^1-9] {
         line_add_char(mu_argcv_unquote_char(yytext[1]));
}

         /* Multi-line strings */
"<<"(-" "?)?\\?{IDENT}[ \t]*.*\n |
"<<"(-" "?)?'{IDENT}'[ \t]*.*\n {
        char *p;

        ++locus.line;
        char_to_strip = NULL;
        multiline_unescape = 1;

        line_begin();
        p = yytext + 2;
        if (*p == '-') {
            ++p;
            if (*p == ' ') {
                ++p;
                char_to_strip = is_space;
            } else
                char_to_strip = is_tab;
        }
        
        if (*p == '\\') {
            p++;
            multiline_unescape = 0;
        }
        if (*p == '\'') {
            char *q;
 
            p++;
            multiline_unescape = 0;
            q = strchr(p, '\'');
            multiline_delimiter_len = q - p;
        } else
            multiline_delimiter_len = strcspn(p, " \t");

        multiline_delimiter = xmalloc(multiline_delimiter_len + 1);
        memcpy(multiline_delimiter, p, multiline_delimiter_len);
        multiline_delimiter[multiline_delimiter_len] = 0;
        if (multiline_unescape)
            BEGIN(ML);
        else
            BEGIN(QML);
}
        /* Quoted multilines */
<QML>[^\\\n]*\n {
        char *p;

        ++locus.line;
        p = yytext;
        if (char_to_strip)
               for (; char_to_strip (*p); p++)
                      ;

        if (strlen(p) >= multiline_delimiter_len
            && memcmp(p, multiline_delimiter, multiline_delimiter_len) == 0
            && isemptystr(p + multiline_delimiter_len)) {
	       free (multiline_delimiter);
	       multiline_delimiter = NULL;
               multiline_delimiter_len = 0;
	       BEGIN(INITIAL);
               line_finish();
               return STRING;
        }
        line_add(p, strlen(p));
}

        /* Unquoted multilines */
<ML>[^\\$%\n]+/[\\$%] {
        char *p = yytext;
        if (char_to_strip)
               for (; char_to_strip (*p); p++)
                      ;
        line_add(p, strlen(p));
        line_finish();
        BEGIN(CML);
        return STRING;
}
<CML>[^\\$%\n]+/[\\$%] {
        line_add(yytext, yyleng);
        line_finish();
        line_begin();
        return STRING;
}
<CML,STR>"%%"|"$$" {
        line_add(yytext, yyleng);
        line_finish();
        return STRING;
}
<ML,CML,STR>[$%] {
        line_add(yytext, yyleng);
}
<ML>[^\\$%\n]*\n {
        char *p;

        ++locus.line;
        p = yytext;
        if (char_to_strip)
               for (; char_to_strip (*p); p++)
                      ;

        if (strlen(p) >= multiline_delimiter_len
            && memcmp(p, multiline_delimiter, multiline_delimiter_len) == 0
            && isemptystr(p + multiline_delimiter_len)) {
	       free (multiline_delimiter);
	       multiline_delimiter = NULL;
               multiline_delimiter_len = 0;
	       BEGIN(INITIAL);
               line_finish();
               return STRING;
        }
        line_add(p, strlen(p));
}
<CML>[^\\$%\n]*\n {
        if (yyleng >= multiline_delimiter_len
            && memcmp(yytext, multiline_delimiter,
                      multiline_delimiter_len) == 0
            && isemptystr(yytext + multiline_delimiter_len)) {
	       free (multiline_delimiter);
	       multiline_delimiter = NULL;
               multiline_delimiter_len = 0;
	       BEGIN(INITIAL);
               line_finish();
               return STRING;
        }
        line_add(yytext, yyleng);
        BEGIN(ML);
}
         /* Other tokens */
{WS}     ;
\n { locus.line++; }
"="|"==" return keyword(EQ);
"!=" return keyword(NE);
"<"  return keyword(LT);
"<=" return keyword(LE);
">"  return keyword(GT);
">=" return keyword(GE);
"&"  return keyword(LOGAND);
"|"  return keyword(LOGOR);
"^"  return keyword(LOGXOR);
"~"  return keyword(LOGNOT);
. return yytext[0];

%%

void
init_string_space()
{
	obstack_init(&string_stk);
}

void
free_string_space()
{
	obstack_free(&string_stk, NULL);
}

char *
mf_strdup(const char *str)
{
	string_add(str, strlen(str) + 1);
	return obstack_finish(&string_stk);
}

struct literal *
string_alloc(const char *str, size_t len)
{
	string_begin();
	string_add(str, len);
	return string_finish();
}

static void
string(const char *str, size_t len)
{
	yylval.literal = string_alloc(str, len);
}

void
string_begin()
{
	/* nothing */
}

struct literal *
string_finish()
{
	char *ptr;
	struct literal *lit;
	obstack_1grow(&string_stk, 0);
	ptr = obstack_finish(&string_stk);
	lit = literal_lookup(ptr);
	if (lit->text != ptr)
		obstack_free(&string_stk, ptr);
	return lit;
}

static void
line_finish()
{	
	yylval.literal = string_finish();
	if (yy_flex_debug)
		fprintf(stderr, "constructed line: %s\n",
			yylval.literal->text);
}		

void
string_add(const char *str, size_t len)
{
	obstack_grow(&string_stk, str, len);
}

void
string_add_char(unsigned char c)
{
	obstack_1grow(&string_stk, c);
}

static void
print_parse_message(const struct locus *locus,
		    const char *prefix, const char *fmt, va_list ap)
{
	int n;
	char buf[512]; /* FIXME */

	if (locus && locus->file)
		n = snprintf(buf, sizeof buf, "%s:%lu: ",
			     locus->file, (unsigned long) locus->line);
	else
		n = 0;
	if (prefix) 
		n += snprintf(buf + n, sizeof buf - n, "%s: ", prefix);
	vsnprintf(buf + n, sizeof buf - n, gettext(fmt), ap);
	mu_error(buf);
}

void
parse_warning(char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	print_parse_message(&locus, _("Warning"), fmt, ap);
	va_end(ap);
}

void
parse_warning_locus(const struct locus *loc, char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	print_parse_message(loc, _("Warning"), fmt, ap);
	va_end(ap);
}

void
parse_error(char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	print_parse_message(&locus, NULL, fmt, ap);
	va_end(ap);
	error_count++;
}

void
parse_error_locus(const struct locus *loc, char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	print_parse_message(loc, NULL, fmt, ap);
	va_end(ap);
	error_count++;
}

int
source(char *name, int ldebug)
{
	yy_flex_debug = ldebug;
	return push_source(name, 0);
}

int
yywrap()
{
	return pop_source();
}

void
add_include_dir(char *dir)
{
	int rc;
	
	if (!include_path) {
		rc = mu_list_create(&include_path);
		if (rc) {
			mu_error(_("Cannot create include path list: %s"),
				 mu_strerror(rc));
			return;
		}
	}
	rc = mu_list_append(include_path, dir);
	if (rc) 
		mu_error(_("Cannot append to the include path list: %s"),
			 mu_strerror(rc));
}

static int
isemptystr(char *text)
{
	for (; *text && isspace (*text); text++)
		;
	return *text == 0;
}

void
onblock(int enable)
{
	if (enable)
		BEGIN(ONBLOCK);
	else
		BEGIN(INITIAL);
}

static int
builtin_macro(const char *s)
{
	if (strcmp(s, "__file__") == 0) {
		yylval.literal = literal_lookup(locus.file);
		return STRING;
	} else if (strcmp(s, "__line__") == 0) {
		yylval.number = locus.line;
		return NUMBER;
	} else if (strcmp(s, "__function__") == 0) {
		s = function_name();
		string(s, strlen(s));
		return STRING;
	} else if (strcmp(s, "__package__") == 0) {
		string(PACKAGE_TARNAME, sizeof PACKAGE_TARNAME - 1);
		return STRING;
	} else if (strcmp(s, "__version__") == 0) {
		string(PACKAGE_VERSION, sizeof PACKAGE_VERSION - 1);
		return STRING;
	} else if (strcmp(s, "__major__") == 0) {
		yylval.number = MAILFROMD_VERSION_MAJOR;
		return NUMBER;
	} else if (strcmp(s, "__minor__") == 0) {
		yylval.number = MAILFROMD_VERSION_MINOR;
		return NUMBER;
	} else if (strcmp(s, "__patch__") == 0) {
		yylval.number = MAILFROMD_VERSION_PATCH;
		return NUMBER;
	} else if (strcmp(s, "__statedir__") == 0) {
		string(DEFAULT_STATE_DIR, sizeof DEFAULT_STATE_DIR - 1);
		return STRING;
	}
	abort();
}

/* End of lex.l */
